
#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from witness.core.abstract import AbstractExtractor
from datetime import datetime
import pandas as pd
import logging

log = logging.getLogger(__name__)


class PandasExtractor(AbstractExtractor):
    """
    Basic pandas extractor class.
    Provides a single 'unify' method for all child pandas extractors.
    """
    def __init__(self, uri):
        super().__init__(uri)

    output: pd.DataFrame

    def _set_extraction_timestamp(self):
        setattr(self, 'extraction_timestamp', datetime.now())

    def extract(self):
        self._set_extraction_timestamp()

    def unify(self):

        data = self.output.to_dict(orient='records')
        meta = {'extraction_timestamp': self.extraction_timestamp,
                'record_source': self.uri}

        setattr(self, 'output', {'meta': meta, 'data': data})

        return self


class PandasFeatherExtractor(PandasExtractor):

    def extract(self):
        df = pd.read_feather(self.uri)
        setattr(self, 'output', df)
        super().extract()

        return self


class PandasExcelExtractor(PandasExtractor):

    def __init__(self, uri, sheet_name=0, header=0, dtype=None):
        self.sheet_name: str or int or None = sheet_name
        self.header: int = header
        self.dtype: str or dict or None = dtype
        super().__init__(uri)

    def extract(self):
        df = pd.read_excel(self.uri,
                           sheet_name=self.sheet_name,
                           header=self.header,
                           dtype=self.dtype)
        setattr(self, 'output', df)
        super().extract()

        return self




#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from .core.batch import Batch
from . import version

# Meta
__version__ = version.version
__author__ = 'Eugene Popov'


import logging
logging.getLogger('witness').addHandler(logging.NullHandler())
formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')



#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.


from witness.providers.pandas.core import PandasLoader


class PandasSQLLoader(PandasLoader):
    """
    Loader that uses Pandas DataFrame.to_sql method for loading data.

    :param engine: sqlalchemy engine;
    :param table: name of the destination table;
    :param schema: name of the destination schema, None if not defined.
    """
    def __init__(self,
                 engine,
                 table: str,
                 schema: str or None = None,
                 method: str or None = None):

        self.engine = engine
        self.schema = schema
        self.table = table
        self.method = method
        uri = f'{schema}.{table}'
        super().__init__(uri)

    def load(self):
        self.output.to_sql(name=self.table,
                           con=self.engine,
                           schema=self.schema,
                           if_exists='append',
                           method=self.method)
        return self


class PandasExcelLoader(PandasLoader):

    def __init__(self, uri, sheet_name='Sheet1'):
        self.sheet_name = sheet_name
        super().__init__(uri)

    def load(self):
        self.output.to_excel(
            excel_writer=self.uri,
            sheet_name=self.sheet_name
        )
        return self


class PandasFeatherLoader(PandasLoader):

    def __init__(self, uri):
        super().__init__(uri)

    def load(self):
        self.output.to_feather(self.uri)
        return self



#
#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

__all__ = ['version']

try:
    import importlib_metadata as metadata
except ImportError:
    from importlib import metadata  # type: ignore[no-redef]

try:
    version = metadata.version('witness-etl')
except metadata.PackageNotFoundError:
    import logging

    # log = logging.getLogger(__name__)
    # log.warning("Package metadata could not be found. Overriding it hardcoded.")
    version = '0.0.1'
    # from setup import version

del metadata




#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
from datetime import datetime

from witness.core.abstract import AbstractExtractor


class FileExtractor(AbstractExtractor):

    def _set_extraction_timestamp(self):
        setattr(self, 'extraction_timestamp', datetime.now())

    def extract(self):
        self._set_extraction_timestamp()

    def unify(self):
        raise NotImplementedError




#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import pandas as pd
import logging
import pickle
from witness.core.abstract import AbstractBatch, AbstractLoader, AbstractExtractor

log = logging.getLogger(__name__)


class PandasBatch(AbstractBatch):

    __slots__ = ('data', 'meta')

    def __init__(self, data=None, meta=None):
        self.data: pd.DataFrame or None = data
        self.meta: dict or None = meta
        self.is_restored = False

    def fill(self, extractor):
        """
        Fills batch internal datastructures using
        the extractor passed in.
        """
        output = extractor.extract().unify().output
        setattr(self, 'data', output['data'])
        setattr(self, 'meta', output['meta'])
        return self

    def push(self, loader, meta_elements: [list[str]] or None = None):
        """
        Pushes data, with the appropriate meta attached,
        to the store defined by the loader passed in.
        """
        loader.prepare(self).attach_meta(meta_elements).load()
        return self

    def _register_dump(self, uri):
        self.meta['dump_uri'] = uri

    def dump(self, uri):
        with open(uri, 'wb') as file:
            pickle.dump(self.data, file)
        self._register_dump(uri)

    def restore(self, uri=None):
        """
        Fills batch with data from dump.
        If no dump uri provided it'll try search in batch meta.
        """
        uri = self.meta['dump_uri'] if uri is None else uri
        with open(uri, 'rb') as file:
            output = pickle.load(file)
        setattr(self, 'data', output)
        self.is_restored = True
        return self

    def info(self):
        pass


class PandasLoader(AbstractLoader):

    def __init__(self, uri):
        super().__init__(uri)

    def prepare(self, batch):
        super().prepare(batch)
        df = pd.DataFrame(batch.data, dtype='str')
        self.output = df
        return self

    def attach_meta(self, meta_elements: [list[str]] or None = None):
        try:
            meta = self.batch.meta
            for element in meta:
                meta[element] = str(meta[element])
        except AttributeError:
            log.exception('No batch object was passed to loader.'
                          'Pass a batch object to "prepare" method first.')
            raise AttributeError('No batch object was passed to loader')
        if meta_elements is None:
            for element in meta:
                self.output[element] = meta[element]
        else:
            for element in meta_elements:
                self.output[element] = meta[element]

        return self

    def load(self):
        raise NotImplementedError




#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
from datetime import datetime
from witness.core.abstract import AbstractExtractor


def is_select(q: str) -> bool:
    return True if 'select' in str.lower(q) else False


class DatabaseExtractor(AbstractExtractor):

    def __init__(self, uri):
        super().__init__(uri)

    def _set_extraction_timestamp(self):
        setattr(self, 'extraction_timestamp', datetime.now())

    def extract(self):
        self._set_extraction_timestamp()

    def unify(self):
        raise NotImplementedError


class ODBCExtractor(DatabaseExtractor):

    def __init__(self, uri: str, query: str):
        self.uri: str = uri
        self.query: str = query if isinstance(query, str) and is_select(query) else None
        super().__init__(uri)

    def _set_extraction_timestamp(self):
        from datetime import datetime
        setattr(self, 'extraction_timestamp', datetime.now())

    def extract(self):
        from pyodbc import connect

        connector = connect(self.uri)
        cursor = connector.cursor()
        rows = cursor.execute(self.query).fetchall()

        self.output = {'description': cursor.description, 'rows': rows}
        super().extract()

        return self

    def unify(self):

        description = self.output['description']
        rows = self.output['rows']
        col_names = [col[0] for col in description]
        dtypes = [col[1] for col in description]
        source_dtypes = {col_names[i]: dtype for i, dtype in enumerate(dtypes)}

        def build_record(names, row):
            record = {names[i]: value for i, value in enumerate(row)}
            return record

        data = [build_record(col_names, row) for row in rows]
        meta = {'extraction_timestamp': self.extraction_timestamp,
                'record_source': self.uri,
                'source_dtypes': source_dtypes}

        setattr(self, 'output', {'meta': meta, 'data': data})

        return self



#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from witness.core.abstract import AbstractBatch
import pickle


class Batch(AbstractBatch):
    """
    Central class of entire lib.
    Able to store standardized data structure
    containing data in form of records and metadata dictionary.
    """

    __slots__ = ('data', 'meta')

    def __init__(self, data=None, meta=None):

        self.data: list or None = data
        self.meta: dict or None = meta
        self.is_restored = False

    def info(self):

        if self.meta is None and self.data is None:
            return 'Batch object is not containing any data.'

        number_of_records = len(self.data) if self.data is not None else None

        message = f"""
        Number of records: {number_of_records}
        Was {'restored from dump ' + f"{self.meta['dump_uri']}" if self.is_restored else 'originally extracted'}
        Source: {self.meta['record_source']}
        Extraction datetime: {self.meta['extraction_timestamp']}
        """

        try:
            message = message + f"Tags: {self.meta['tags']}\n"
        except KeyError:
            pass

        return message

    def fill(self, extractor):
        """
        Fills batch internal datastructures using
        the extractor passed in.
        """
        output = extractor.extract().unify().output
        setattr(self, 'data', output['data'])
        setattr(self, 'meta', output['meta'])
        return self

    def push(self, loader, meta_elements: [list[str]] or None = None):
        """
        Pushes data, with the appropriate meta attached,
        to the store defined by the loader passed in.
        """
        loader.prepare(self).attach_meta(meta_elements).load()
        return self

    def _register_dump(self, uri):
        self.meta['dump_uri'] = uri

    def dump(self, uri):
        """
        Dumps batch data to pickle file with defined uri.
        """
        with open(uri, 'wb') as file:
            pickle.dump(self.data, file)
        self._register_dump(uri)

    def restore(self, uri=None):
        """
        Fills batch with data from dump.
        If no dump uri provided it'll try search in batch meta.
        """
        uri = self.meta['dump_uri'] if uri is None else uri
        with open(uri, 'rb') as file:
            output = pickle.load(file)
        setattr(self, 'data', output)
        self.is_restored = True
        return self




#  Copyright (c) 2022.  Eugene Popov.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import logging
from abc import ABCMeta, abstractmethod
from datetime import datetime

log = logging.getLogger(__name__)


class AbstractBatch(metaclass=ABCMeta):

    @abstractmethod
    def fill(self, extractor):
        raise NotImplemented

    @abstractmethod
    def push(self, loader):
        raise NotImplemented

    @abstractmethod
    def dump(self, uri):
        raise NotImplemented

    @abstractmethod
    def restore(self, uri):
        raise NotImplemented


class AbstractExtractor(metaclass=ABCMeta):

    def __init__(self, uri=None):

        self.uri = uri
        self.output = None
        self.extraction_timestamp: datetime or None = None

    @abstractmethod
    def _set_extraction_timestamp(self):
        raise NotImplementedError

    @abstractmethod
    def extract(self):
        """
        An abstract method for data extraction.
        """
        raise NotImplementedError

    @abstractmethod
    def unify(self):
        """
        An abstract method for deserialization from data source.
        """
        raise NotImplementedError


class AbstractLoader(metaclass=ABCMeta):

    def __init__(self, uri=None):

        self.uri = uri
        self.batch = None
        self.output = None

    @abstractmethod
    def prepare(self, batch):
        """
        An abstract method of preparing data from a Batch object for loading.
        """
        self._set_batch(batch)

    @abstractmethod
    def attach_meta(self, meta_elements: [list[str]] or None = None):
        """
        An abstract method for attaching meta from Batch-object
        to data prepared for loading.
        """
        raise NotImplementedError

    @abstractmethod
    def load(self):
        """
        An abstract method for loading data to destination store.
        """
        raise NotImplementedError

    def _set_batch(self, batch):
        setattr(self, 'batch', batch)




from datetime import datetime
from witness.core.abstract import AbstractExtractor
import requests
from requests.auth import AuthBase


class HttpGetExtractor(AbstractExtractor):

    def __init__(self, uri, params: dict or None = None, auth: AuthBase = None):
        self.params: dict or None = params
        self.auth = auth
        super().__init__(uri)

    def _set_extraction_timestamp(self):
        setattr(self, 'extraction_timestamp', datetime.now())

    def extract(self):

        response = requests.get(url=self.uri, params=self.params, auth=self.auth)
        response.raise_for_status()

        setattr(self, 'output', response)
        self._set_extraction_timestamp()

        return self

    def unify(self):
        raise NotImplementedError


class JsonHttpGetExtractor(HttpGetExtractor):

    def unify(self):

        data = self.output.json()
        meta = {'extraction_timestamp': self.extraction_timestamp,
                'record_source': self.uri}

        setattr(self, 'output', {'meta': meta, 'data': data})

        return self
